searchData={"content_type":"text/markdown","items":[{"doc":"A generic Plug implementing an Incremental Static Regeneration (ISR) pattern.\n\nThis plug allows serving cached data (fresh or stale) quickly while triggering\nnon-blocking background tasks to refresh expired data. It's designed to be\nflexible and reusable in Phoenix applications.","ref":"ISRPlug.html","title":"ISRPlug","type":"module"},{"doc":"The plug is configured via options passed during the `plug` call in the router:\n\n```elixir\nplug ISRPlug,\n  # --- Required ---\n  fetch_fun: &YourModule.your_fetch_function/1,\n  apply_fun: &YourModule.your_apply_function/2,\n\n  # --- Optional ---\n  extract_data_fun: &YourModule.your_extract_function/1,\n  cache_key_fun: &YourModule.your_cache_key_function/1,\n  ets_table: :my_specific_isr_cache,\n  cache_ttl_ms: 60_000, # 1 minute\n  stale_serving_ttl_ms: 3_600_000, # 1 hour\n  error_handler_fun: &YourModule.your_error_handler/2\n```\n\nSee the function documentation for `init/1` for details on each option.","ref":"ISRPlug.html#module-configuration","title":"Configuration - ISRPlug","type":"module"},{"doc":"Processes the connection according to the ISR logic.\n\n1. Extracts data and determines the cache key using configured functions.\n2. Checks the ETS cache.\n3. Handles cache hits (fresh or stale).\n4. If stale, serves stale data and triggers a background refresh task.\n5. Handles cache misses or expired stale TTL by fetching synchronously.\n6. If synchronous fetch fails, calls the error handler.\n7. If data is available (fresh, stale, or newly fetched), applies it using the configured function.\n8. Returns the (potentially modified) connection.","ref":"ISRPlug.html#call/2","title":"ISRPlug.call/2","type":"function"},{"doc":"Initializes the plug with configuration options.","ref":"ISRPlug.html#init/1","title":"ISRPlug.init/1","type":"function"},{"doc":"* `:fetch_fun` (**required**): A 1-arity function\n    `(extracted_data :: any()) -> {:ok, value :: any()} | {:error, reason :: any()}`.\n    Performs the data retrieval.\n\n  * `:apply_fun` (**required**): A 2-arity function\n    `(conn :: Plug.Conn.t(), value :: any()) -> Plug.Conn.t()`.\n    Applies the successfully retrieved `value` (fresh or stale) to the connection.\n    Must return a `Plug.Conn.t()`.\n\n  * `:extract_data_fun` (*optional*): A 1-arity function\n    `(conn :: Plug.Conn.t()) -> any()`.\n    Extracts necessary data from the `conn` to be passed to `:fetch_fun`.\n    Defaults to `fn _conn -> %{} end`.\n\n  * `:cache_key_fun` (*optional*): A 1-arity function\n    `(conn :: Plug.Conn.t()) -> term()`.\n    Generates a unique ETS key based on the connection.\n    Defaults to `fn _conn -> :isr_plug_default_key end`. Ensure this generates\n    distinct keys if the fetched data varies based on connection properties.\n\n  * `:ets_table` (*optional*): Atom name for the ETS table.\n    Defaults to `isr_plug_cache`. Use distinct names if using the plug\n    multiple times for different purposes.\n\n  * `:cache_ttl_ms` (*optional*): Integer milliseconds for how long data is\n    considered fresh. Defaults to `60000` (1 minute).\n\n  * `:stale_serving_ttl_ms` (*optional*): Integer milliseconds *after* expiry\n    during which stale data can be served while refreshing.\n    Defaults to `3600000` (1 hour).\n\n  * `:error_handler_fun` (*optional*): A 2-arity function\n    `(conn :: Plug.Conn.t(), reason :: any()) -> Plug.Conn.t()`.\n    Called only when a *synchronous* fetch fails. Defaults to a function that\n    logs the error and returns the original `conn`.\n\nThis function also initializes the specified ETS table if it doesn't already exist.","ref":"ISRPlug.html#init/1-options","title":"Options - ISRPlug.init/1","type":"function"},{"doc":"# ISRPlug\n\n[![Hex.pm](https://img.shields.io/hexpm/v/isr_plug.svg)](https://hex.pm/packages/isr_plug)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/isr_plug/)\n\n`ISRPlug` is a generic, reusable Elixir Plug designed to implement the Incremental Static Regeneration (ISR) pattern within Phoenix (or other Plug-based) applications.\n\nIt helps improve performance and maintain data freshness by:\n\n*   Serving cached data (fresh or slightly stale) quickly.\n*   Triggering non-blocking background tasks to refresh expired data.\n*   Allowing flexible, application-specific logic for fetching data and applying it to the `Plug.Conn`.","ref":"readme.html","title":"ISRPlug","type":"extras"},{"doc":"Web applications often need dynamic data (e.g., configuration, feature flags, content) per request. Fetching this synchronously can slow down responses. Standard caching helps but can lead to serving stale data indefinitely or requiring complex invalidation.\n\nISR offers a balance: serve stale data for a limited time while refreshing in the background, ensuring eventual consistency without blocking the user request.","ref":"readme.html#problem-addressed","title":"Problem Addressed - ISRPlug","type":"extras"},{"doc":"Add `isr_plug` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:isr_plug, \"~> 0.1.0\"} # Or use path: {:isr_plug, path: \"../path/to/isr_plug\"} for local dev\n    # Or git: {:isr_plug, git: \"https://github.com/your_github_username/isr_plug.git\", tag: \"v0.1.0\"}\n  ]\nend\n```\n\nThen, run `mix deps.get`.","ref":"readme.html#installation","title":"Installation - ISRPlug","type":"extras"},{"doc":"1.  **Define Your Logic Module:** Create a module in your application containing the functions needed by the plug.\n\n    ```elixir\n    # lib/my_app/dynamic_config.ex\n    defmodule MyApp.DynamicConfig do\n      require Logger\n\n      # --- FETCH FUNCTION ---\n      # Fetches the data. Receives data extracted by `extract_data_fun`.\n      # Must return {:ok, value} or {:error, reason}.\n      def fetch_settings(extracted_data) do\n        tenant_id = extracted_data[:tenant_id]\n        Logger.info(\"Fetching settings for tenant: #{tenant_id}\")\n        # Simulate fetching from DB or API\n        settings = %{feature_x_enabled: true, theme: \"dark\"} # Example data\n        {:ok, settings}\n        # Or: {:error, :database_timeout}\n      end\n\n      # --- APPLY FUNCTION ---\n      # Applies the fetched data to the connection.\n      # Receives the conn and the fetched value. Must return the conn.\n      def apply_settings(conn, settings) do\n        Plug.Conn.assign(conn, :dynamic_settings, settings)\n      end\n\n      # --- (Optional) EXTRACT DATA FUNCTION ---\n      # Extracts data needed for fetching from the conn.\n      # Defaults to fn _ -> %{} end.\n      def extract_tenant_id(conn) do\n         # Example: Get tenant from subdomain or session\n         %{tenant_id: conn.host |> String.split(\".\") |> List.first()}\n      end\n\n      # --- (Optional) CACHE KEY FUNCTION ---\n      # Generates a unique cache key based on the conn.\n      # Defaults to fn _ -> :isr_plug_default_key end.\n      # IMPORTANT: Use if fetched data varies per request context (user, tenant, etc.)\n      def settings_cache_key(conn) do\n        tenant_id = conn.host |> String.split(\".\") |> List.first()\n        {:settings, tenant_id}\n      end\n\n      # --- (Optional) ERROR HANDLER FUNCTION ---\n      # Handles errors during *synchronous* fetches (cache miss/expired stale).\n      # Receives conn and the error reason. Must return the conn.\n      # Default logs the error and returns the conn unchanged.\n      def handle_fetch_error(conn, reason) do\n        Logger.error(\"Failed to fetch dynamic settings: #{inspect reason}. Assigning defaults.\")\n        # Example: Assign default settings or halt\n        Plug.Conn.assign(conn, :dynamic_settings, %{feature_x_enabled: false, theme: \"light\"})\n        # Or: conn |> Plug.Conn.send_resp(:internal_server_error, \"Config error\") |> Plug.Conn.halt()\n      end\n    end\n    ```\n\n2.  **Add the Plug to Your Pipeline:** In your `router.ex` or `endpoint.ex`, add `ISRPlug` with your configuration.\n\n    ```elixir\n    # lib/my_app_web/router.ex\n    defmodule MyAppWeb.Router do\n      use MyAppWeb, :router\n      import ISRPlug # Import to use `plug ISRPlug` directly\n      alias MyApp.DynamicConfig\n\n      pipeline :browser do\n        plug :accepts, [\"html\"]\n        # ... other plugs ...\n\n        # Add ISRPlug to fetch dynamic settings\n        plug ISRPlug,\n          fetch_fun: &DynamicConfig.fetch_settings/1,\n          apply_fun: &DynamicConfig.apply_settings/2,\n          # Optional functions:\n          extract_data_fun: &DynamicConfig.extract_tenant_id/1,\n          cache_key_fun: &DynamicConfig.settings_cache_key/1,\n          error_handler_fun: &DynamicConfig.handle_fetch_error/2,\n          # Optional configuration:\n          ets_table: :dynamic_settings_cache, # Use a specific ETS table name\n          cache_ttl_ms: :timer.minutes(5),    # Data is fresh for 5 minutes\n          stale_serving_ttl_ms: :timer.hours(1) # Serve stale for 1 hour while refreshing\n\n        # ... other plugs ...\n      end\n\n      scope \"/\", MyAppWeb do\n        pipe_through :browser\n        get \"/\", PageController, :index\n      end\n    end\n    ```","ref":"readme.html#usage","title":"Usage - ISRPlug","type":"extras"},{"doc":"See `ISRPlug.init/1` documentation for details on all options:\n\n*   `:fetch_fun` (**required**)\n*   `:apply_fun` (**required**)\n*   `:extract_data_fun` (*optional*)\n*   `:cache_key_fun` (*optional*)\n*   `:ets_table` (*optional*, defaults to `:isr_plug_cache`)\n*   `:cache_ttl_ms` (*optional*, defaults to 1 minute)\n*   `:stale_serving_ttl_ms` (*optional*, defaults to 1 hour)\n*   `:error_handler_fun` (*optional*)","ref":"readme.html#configuration-options","title":"Configuration Options - ISRPlug","type":"extras"},{"doc":"1.  On request, `extract_data_fun` runs, then `cache_key_fun` determines the ETS cache key.\n2.  The ETS table (`:ets_table`) is checked for the key.\n3.  **Cache Hit (Fresh):** If data exists and hasn't passed `cache_ttl_ms`, `apply_fun` runs with the cached value.\n4.  **Cache Hit (Stale):** If data exists, passed `cache_ttl_ms`, but *not* `stale_serving_ttl_ms`, `apply_fun` runs with the *stale* cached value, AND a non-blocking `Task` starts in the background to call `fetch_fun` and update the cache.\n5.  **Cache Miss / Too Stale:** If no data exists, or it passed `stale_serving_ttl_ms`, `fetch_fun` runs *synchronously*.\n    *   **Success:** The result is cached, and `apply_fun` runs with the new value.\n    *   **Failure:** `error_handler_fun` runs. The default handler logs the error and passes the connection through; `apply_fun` is *not* called.\n6.  The potentially modified connection is passed to the next plug.","ref":"readme.html#how-it-works","title":"How it Works - ISRPlug","type":"extras"},{"doc":"","ref":"readme.html#testing-the-plug","title":"Testing the Plug - ISRPlug","type":"extras"},{"doc":"The library includes unit tests. Clone the repository and run:\n\n```bash\nmix deps.get\nmix test\n```","ref":"readme.html#running-unit-tests","title":"Running Unit Tests - ISRPlug","type":"extras"},{"doc":"1.  **Add as Path Dependency:** In your *consuming application's* `mix.exs`, add `isr_plug` using a path dependency:\n\n    ```elixir\n    # In your Phoenix app's mix.exs\n    def deps do\n      [\n        # ... other deps\n        {:isr_plug, path: \"../path/to/your/local/isr_plug/checkout\"}\n      ]\n    end\n    ```\n\n2.  **Run `mix deps.get`** in your consuming application.\n3.  **Configure the Plug:** Add the plug to your router or endpoint as shown in the [Usage](#usage) section.\n4.  **Observe Logs:** Start your Phoenix server (`mix phx.server`). Make requests to the relevant endpoints. Check your application logs for messages from `[ISRPlug]` indicating cache hits (fresh/stale), misses, synchronous fetches, and background refreshes.\n5.  **Verify Behavior:**\n    *   Check if the data applied by your `apply_fun` (e.g., assigns, headers) is present in the `conn` or response.\n    *   Modify the underlying data source that your `fetch_fun` uses. Observe if the application picks up the change after the `cache_ttl_ms` + `stale_serving_ttl_ms` duration (on the next request after that period), or sooner if a background refresh completes after the `cache_ttl_ms`.\n    *   Simulate errors in your `fetch_fun` and verify that your `error_handler_fun` is called correctly during synchronous fetches.","ref":"readme.html#testing-integration-in-your-application-development","title":"Testing Integration in Your Application (Development) - ISRPlug","type":"extras"},{"doc":"The current implementation of `ISRPlug` uses ETS for caching. When stale data is encountered, it triggers a background `Task` to refresh the data.\n\n**Potential Issue: Concurrent Background Refreshes**\n\nPhoenix handles requests using multiple concurrent processes. ETS provides shared memory on a single node, and operations like checking the cache or updating it are generally safe. However, there's a subtle **race condition** specifically related to triggering the **background refresh task**:\n\n1.  **Request A (Process A):** Checks the cache, finds stale data, decides a background refresh is needed.\n2.  **Request B (Process B):** *Almost simultaneously*, checks the cache, finds the *same* stale data, and *also* decides a background refresh is needed.\n3.  **Outcome:** Both Process A and Process B might independently start a background `Task` to execute your `:fetch_fun` before either task has had a chance to update the cache.\n\nThis can lead to:\n\n*   **Redundant Work:** Your potentially expensive `:fetch_fun` (DB query, API call) runs multiple times when only one refresh was necessary.\n*   **Resource Spikes:** More temporary processes (`Task`s) are created than needed.\n*   **Downstream Pressure:** Increased load on databases or external APIs, potentially hitting rate limits.\n\n**Why This Happens (ETS vs. Serialized Logic):**\n\n*   Each request runs in its own process.\n*   The sequence \"Check Cache -> Decide Refresh -> Start Task\" is performed independently by each process.\n*   While ETS operations themselves are fast and reasonably atomic, there's no mechanism *inherent in this Plug's ETS-based approach* to guarantee that only *one* process \"wins\" the right to start the background task during that tiny window between deciding and acting.\n\n**Why Locking is NOT Applied to Synchronous Fetches:**\n\nThe plug **intentionally does not** attempt to prevent concurrent *synchronous* fetches (when data is missing entirely or too stale). Applying a lock here would mean:\n\n1.  Request A misses the cache and starts a synchronous fetch, acquiring a hypothetical lock.\n2.  Request B also misses the cache but would be *blocked*, waiting for Request A to finish *and* release the lock, even though Request B also needs the data immediately.\n3.  This would significantly slow down responses during cache misses, which is generally undesirable. Allowing concurrent synchronous fetches, while slightly redundant, ensures requests aren't blocked waiting for *other* requests to fetch the same missing data. The first one to finish populates the cache.\n\n**Alternative (More Complex) Solution: GenServer Coordinator**\n\nA more robust way to guarantee that only one background refresh task runs at a time (on a single node) involves using a central coordinating process, typically a `GenServer`:\n\n1.  **Coordinator:** A single `GenServer` process manages the state of which cache keys are currently being refreshed.\n2.  **Plug Interaction:** When the `ISRPlug` detects stale data, instead of starting a `Task` directly, it sends a non-blocking message (e.g., `GenServer.cast`) to the Coordinator, requesting a refresh for the specific `cache_key`.\n3.  **Serialized Decision:** The Coordinator processes these requests one by one. It checks its internal state. If a refresh for that key isn't already marked \"in progress,\" it marks it, starts the background `Task`, and configures the Task to notify the Coordinator upon completion (to clear the \"in progress\" state). If a refresh *is* already marked, the Coordinator simply ignores the duplicate request.\n\n**Why We Didn't Implement the GenServer:**\n\n*   **Increased Complexity:** Requires adding and managing a dedicated GenServer process within your application's supervision tree.\n*   **Potential Bottleneck:** While likely negligible for many use cases, the single GenServer becomes a serialization point for *all* background refresh triggers managed by instances of this plug.\n*   **Added Resource:** Introduces another persistent process to the system.\n\n**Current Trade-off:**\n\nThe current `ISRPlug` implementation prioritizes simplicity and avoids the overhead of a dedicated GenServer coordinator. It accepts the possibility of occasional redundant *background* refresh tasks under high concurrent load on stale keys. This is often an acceptable trade-off, especially if the `:fetch_fun` is reasonably fast and idempotent, and downstream rate limits are not a major concern.\n\nIf absolute prevention of concurrent background refreshes is critical for your specific use case (e.g., very expensive fetches, strict rate limits), consider implementing a custom solution using the GenServer coordinator pattern described above.","ref":"readme.html#concurrency-considerations-and-limitations-ets-locking","title":"Concurrency Considerations and Limitations (ETS Locking) - ISRPlug","type":"extras"},{"doc":"Contributions are welcome! Please open an issue or submit a pull request. (TODO: Add contribution guidelines).","ref":"readme.html#contributing","title":"Contributing - ISRPlug","type":"extras"},{"doc":"This project is licensed under the Apache 2.0 License - see the LICENSE file for details. (PLEASE VERIFY/CHANGE and add a LICENSE file).","ref":"readme.html#license","title":"License - ISRPlug","type":"extras"}],"producer":{"name":"ex_doc","version":"0.37.3"},"proglang":"elixir"}